! Input a rotation in the axis-angle representation, and output the Euler angles of the rotation
! 
! n1 = time axis
! n2 = trace number
!
! uvec_dp = the normalized rotation axis
! theta_dp = the angle of rotations around the rotation axis

program axisangle2euler
	use sep
	use quat_mod
	implicit none


	integer, external :: sreed, srite
	integer	:: n1, n2, it, i2, jt
	real	:: d1, o1, d2, o2, d3, o3
	real	:: eps=0., uvecx, uvecy, uvecz
	real(kind=8)	:: eps_dp, idt, denom, cos_theta_dp, test_acos
	real, parameter	:: deg2rad = 4.*atan(1.)/180.
	type(quat_dp) :: q1
	type(quat) :: q_sp

	real, allocatable, dimension (:,:)	:: theta				! first time derivative of the angle of rotation between the magnetic vector's projection
																! on the magnetic components over two consecutive time steps
	real, allocatable, dimension (:,:,:)	:: uvec				! unit vector (rotation axis) around which the rotation took place over two consecutive time samples.
	real, allocatable, dimension (:,:,:)	:: euler_rot		! rotations in terms of euler angles in the Tait-Bryan convention

	real(kind=8), allocatable, dimension (:,:)	:: theta_dp
	real(kind=8), allocatable, dimension (:,:,:)	:: uvec_dp

	call sep_init()

	call from_aux ("theta", "n1", n1)
	call from_aux ("theta", "o1", o1)
	call from_aux ("theta", "d1", d1)
	call from_aux ("theta", "n2", n2, 1)
	call from_aux ("theta", "o2", o2, 0.)
	call from_aux ("theta", "d2", d2, 1.)


	idt = DBLE(1./d1)

	call from_param ('X', uvecx, 1.)
	call from_param ('Y', uvecy, 1.)
	call from_param ('Z', uvecz, 1.)
	call from_param ('jt', jt, 100)
	call from_param ('eps', eps, 1e-20)
	eps_dp = DBLE(eps)

	if (exist_file ("euler")) then
		call to_history ("n1", n1, "euler")
		call to_history ("o1", o1, "euler")
		call to_history ("d1", d1, "euler")
		call to_history ("n2", 3,  "euler")
		call to_history ("o2", 0., "euler")
		call to_history ("d2", 1., "euler")
		call to_history ("n3", n2, "euler")
		call to_history ("o3", o2, "euler")
		call to_history ("d3", d2, "euler")
	endif
	if (exist_file ("qout")) then
		call to_history ("n1", 4,  "qout")
		call to_history ("o1", 0., "qout")
		call to_history ("d1", 1., "qout")
		call to_history ("n2", n1, "qout")
		call to_history ("o2", o1, "qout")
		call to_history ("d2", d1, "qout")
		call to_history ("n3", n2, "qout")
		call to_history ("o3", o2, "qout")
		call to_history ("d3", d2, "qout")
	endif

	allocate (theta(n1,n2),    uvec(n1,3,n2), euler_rot(n1,3,n2))
	allocate (theta_dp(n1,n2), uvec_dp(n1,3,n2))

	theta = 0.; uvec = 0.; theta_dp = 0.; uvec_dp = 0.; euler_rot = 0.

	if (sreed ("theta", theta, 4*n1*n2) .ne. 4*n1*n2) call seperr ("error reading from input file theta")
	theta_dp = theta

	uvec_dp(:,1,:) = uvecx
	uvec_dp(:,2,:) = uvecy
	uvec_dp(:,3,:) = uvecz


	q1 = qnew (DBLE(1.), DBLE(0.), DBLE(0.), DBLE(0.))
	q_sp = qnew (1.,0.,0.,0.)
	if (exist_file ("qout")) then
		call q_sp%quat_write ("qout")
	endif

	! calculate rotation for all time samples other than the first one
	do i2 = 1, n2
	do it = 1, n1
		!write (0,*), " "
		!write (0,*), " "
		!write (0,*), "i2, it:", i2, it

		!write (0,*), "theta = ", theta_dp(it,i2)

		denom = 1. / (sqrt(uvec_dp(it,1,i2)*uvec_dp(it,1,i2) + uvec_dp(it,2,i2)*uvec_dp(it,2,i2) + uvec_dp(it,3,i2)*uvec_dp(it,3,i2)) + eps_dp)
		uvec_dp (it,1,i2) = uvec_dp(it,1,i2) * denom
		uvec_dp (it,2,i2) = uvec_dp(it,2,i2) * denom
		uvec_dp (it,3,i2) = uvec_dp(it,3,i2) * denom

		!write (0,*), "denom = ", denom
		!write (0,*), "nrm", uvec_dp(it,1,i2), uvec_dp(it,2,i2), uvec_dp(it,3,i2)

		! rotate the quaternion using the axis and angle
		call q1%quat_dp_mult_fromangle (theta_dp(it,i2), uvec_dp(it,1,i2), uvec_dp(it,2,i2), uvec_dp(it,3,i2))

		!call q1%quat_dp_print ()

		if (mod(it,jt) == 0) then
			call q1%quat_dp_norm ()
		endif

		if (exist_file ("qout")) then
			call q1%quat_dp2sp (q_sp)
			call q_sp%quat_write ("qout")
		endif

		! convert to an euler angle representation of the quaternion
		call q1%quat_dp2euler (euler_rot(it,1,i2), euler_rot(it,2,i2), euler_rot(it,3,i2))
	end do
	call q_sp%quat_write ("qout")
	end do


	theta = REAL(theta_dp)
	uvec  = REAL(uvec_dp)

	if (exist_file ("euler")) then
		if (srite ("euler", euler_rot, 4*n1*3*n2) .ne. 4*n1*3*n2) call seperr ("error writing to output file euler")
	endif
	deallocate (theta, uvec, euler_rot, theta_dp, uvec_dp)

	call sep_close()

end program axisangle2euler
