! Quaternion module for basic quaternion math operations, for double precision and single precision.
! (The interfaces aren't working for me when I call them from the main program,
! probably because I'm using intel fortran compiler version 12 instead of 14)
!

module quat_mod

	implicit none

	real, private :: tol=1e-12
	real(kind=8), private :: tol_dp=1e-15

	type quat
		private
		real	:: w, x, y, z
		contains
			procedure, nopass, public	:: quat_def_
			procedure, public			:: quat_mult_fromangle
			procedure, public			:: quat_mult_fromquat
			procedure, public			:: quat_norm
			procedure, public			:: quat_print
			procedure, public			:: quat_write
			procedure, public			:: quat2euler
	end type

	type quat_dp
		private
		real(kind=8)	:: w, x, y, z
		contains
			procedure, nopass, public	:: quat_dp_def_
			procedure, public			:: quat_dp_mult_fromangle
			procedure, public			:: quat_dp_mult_fromquat
			procedure, public			:: quat_dp_norm
			procedure, public			:: quat_dp_print
			procedure, public			:: quat_dp_write
			procedure, public			:: quat_dp2sp
			procedure, public			:: quat_dp2euler
	end type

	interface qnew
		module procedure quat_def_, quat_dp_def_
	end interface

	interface qmult
		module procedure quat_mult_fromangle, quat_dp_mult_fromangle, quat_mult_fromquat, quat_dp_mult_fromquat
	end interface

	interface qnorm
		module procedure quat_norm, quat_dp_norm
	end interface

	interface qprint
		module procedure quat_print, quat_dp_print
	end interface

	interface q2euler
		module procedure quat2euler, quat_dp2euler
	end interface

contains

	! quaternion constructor
	function quat_def_ (w_in, x_in, y_in, z_in) result (newquat)
		class(quat), allocatable	:: newquat
		real, intent(in)			:: w_in, x_in, y_in, z_in

		allocate (quat::newquat)
		newquat%w = w_in
		newquat%x = x_in
		newquat%y = y_in
		newquat%z = z_in
	end function

	function quat_dp_def_ (w_in, x_in, y_in, z_in) result (newquat)
		class(quat_dp), allocatable	:: newquat
		real(kind=8), intent(in)	:: w_in, x_in, y_in, z_in

		allocate (quat_dp::newquat)
		newquat%w = w_in
		newquat%x = x_in
		newquat%y = y_in
		newquat%z = z_in
	end function

	! convert from double precision quaternion to single precision.
	subroutine quat_dp2sp (this, qsp)
		class(quat_dp), intent(inout)	:: this
		type(quat), intent(inout)		:: qsp

		qsp.w = real(this.w)
		qsp.x = real(this.x)
		qsp.y = real(this.y)
		qsp.z = real(this.z)
	end subroutine

	! generate a quaternion q from the rotation axis and angle, and then use it to rotate this quaternion.
	subroutine quat_mult_fromangle (this, theta, u1, u2, u3)
		class(quat), intent(inout)	:: this
		real,        intent(in)		:: theta, u1, u2, u3
		type(quat)					:: q2

		q2 = qnew (cos(0.5*theta), u1*sin(0.5*theta), u2*sin(0.5*theta), u3*sin(0.5*theta))
		call this%quat_mult_fromquat (q2)
	end subroutine

	subroutine quat_dp_mult_fromangle (this, theta, u1, u2, u3)
		class(quat_dp), intent(inout)	:: this
		real(kind=8),   intent(in)		:: theta, u1, u2, u3
		type(quat_dp)					:: q2

		q2 = qnew (cos(DBLE(0.5)*theta), u1*sin(DBLE(0.5)*theta), u2*sin(DBLE(0.5)*theta), u3*sin(DBLE(0.5)*theta))
		call this%quat_dp_mult_fromquat (q2)
	end subroutine

	! multiply this quaternion by q, which equates to rotating by q.
	subroutine quat_mult_fromquat (this, q)
		class(quat), intent(inout)	:: this
		type(quat),  intent(in)		:: q

		this%w = q%w*this%w - q%x*this%x - q%y*this%y - q%z*this%z
		this%x = q%w*this%x + q%x*this%w + q%y*this%z - q%z*this%y
		this%y = q%w*this%y - q%x*this%z + q%y*this%w + q%z*this%x
		this%z = q%w*this%z + q%x*this%y - q%y*this%x + q%z*this%w
	end subroutine

	subroutine quat_dp_mult_fromquat (this, q)
		class(quat_dp), intent(inout)	:: this
		type(quat_dp),  intent(in)		:: q

		this%w = q%w*this%w - q%x*this%x - q%y*this%y - q%z*this%z
		this%x = q%w*this%x + q%x*this%w + q%y*this%z - q%z*this%y
		this%y = q%w*this%y - q%x*this%z + q%y*this%w + q%z*this%x
		this%z = q%w*this%z + q%x*this%y - q%y*this%x + q%z*this%w
	end subroutine

	! normalize the quaternion so that |q| = 1.
	subroutine quat_norm (this)
		class(quat), intent(inout)	:: this
		real	:: norm

		norm = sqrt(this.w**2 + this.x**2 + this.y**2 + this.z**2)
		if (abs(norm - 1.) .gt. tol) then
			norm = DBLE(1.) / norm
			this.w = this.w * norm
			this.x = this.x * norm
			this.y = this.y * norm
			this.z = this.z * norm
		endif
	end subroutine

	subroutine quat_dp_norm (this)
		class(quat_dp), intent(inout)	:: this
		real(kind=8)	:: norm

		norm = sqrt(this.w**2 + this.x**2 + this.y**2 + this.z**2)
		if (abs(norm - DBLE(1.)) .gt. tol_dp) then
			norm = DBLE(1.) / norm
			this.w = this.w * norm
			this.x = this.x * norm
			this.y = this.y * norm
			this.z = this.z * norm
		endif
	end subroutine

	! convert the quaternion to the Tait-Bryan Euler angle representation.
	! R = Rz(psi) Ry(theta) Rx(phi).
	! This is a left-hand rule apparently.
	! Source: Paul Berner, Orientation, Rotation, Velocity and Acceleration, and the SRM,
	! Version 2.0, June 2008, SEDRIS
	subroutine quat2euler (this, phi, theta, psi)
		class(quat), intent(inout)	:: this
		real,        intent(out)	:: phi, theta, psi

		phi   = REAL (atan2 (2.*this%y*this%z + 2.*this%w*this%x, (this%z*this%z - this%y*this%y - this%x*this%x + this%w*this%w)))
		theta = REAL (-asin  (2.*(this%x*this%z - this%w*this%y)))
		psi   = REAL (atan2 (2.*this%x*this%y + this%w*this%z, (this%x*this%x + this%w*this%w - this%z*this%z - this%y*this%y)))
	end subroutine

	subroutine quat_dp2euler (this, phi, theta, psi)
		class(quat_dp), intent(inout)	:: this
		real,           intent(out)		:: phi, theta, psi

		phi   = REAL (atan2 (2.*this%y*this%z + 2.*this%w*this%x, (this%z*this%z - this%y*this%y - this%x*this%x + this%w*this%w)))
		theta = REAL (-asin  (2.*(this%x*this%z - this%w*this%y)))
		psi   = REAL (atan2 (2.*this%x*this%y + this%w*this%z, (this%x*this%x + this%w*this%w - this%z*this%z - this%y*this%y)))
	end subroutine

	! print quaternion to screen
	subroutine quat_print (this)
		class(quat), intent(in)	:: this
		write (0, '(f12.9,a,f12.9,a,f12.9,a,f12.9)'), this.w, ":", this.x, ":", this.y, ":", this.z
	end subroutine

	subroutine quat_dp_print (this)
		class(quat_dp), intent(in)	:: this
		write (0, '(f15.12,a,f15.12,a,f15.12,a,f15.12)'), this.w, ":", this.x, ":", this.y, ":", this.z
	end subroutine

	! write quaternion to SEP history file
	subroutine quat_write (this, tag)
		class(quat), intent(in)			:: this
		character(len=*), intent(in)	:: tag
		real							:: tmp(4)
		integer, external				:: srite

		tmp(1) = this.w; tmp(2) = this.x; tmp(3) = this.y; tmp(4) = this.z
		if (srite (tag, tmp, 4*4) .ne. 4*4) then
			write (0,*), "tag:", trim(tag)
			call seperr ("quat_mod: error writing out quaternion to file")
		endif
	end subroutine

	subroutine quat_dp_write (this, tag)
		class(quat_dp), intent(in)		:: this
		character(len=*), intent(in)	:: tag
		real(kind=8)					:: tmp(4)
		integer, external				:: srite

		tmp(1) = this.w; tmp(2) = this.x; tmp(3) = this.y; tmp(4) = this.z
		if (srite (tag, tmp, 8*4) .ne. 8*4) then
			write (0,*), "tag:", trim(tag)
			call seperr ("quat_mod: error writing out quaternion to file")
		endif
	end subroutine
end module quat_mod
